@@ -2,12 +2,12 @@
  * load_dmf.cpp
  * ------------
  * Purpose: DMF module loader (X-Tracker by D-LUSiON).
- * Notes  : If it wasn't already outdated when the tracker was released, this would be a rather interesting
+ * Notes  : If it wasn't already outdated when the tracker left beta state, this would be a rather interesting
  *          and in some parts even sophisticated format - effect columns are separated by effect type, an easy to
  *          understand BPM tempo mode, effect durations are always divided into a 256th row, vibrato effects are
  *          specified by period length and the same 8-Bit granularity is used for both volume and panning.
  *          Unluckily, this format does not offer any envelopes or multi-sample instruments, and bidi sample loops
- *          are missing as well, so it was already well behind FT2 and IT back then.
+ *          are missing as well, so it was already well behind FT2 back then.
  * Authors: Johannes Schultz (mostly based on DMF.TXT, DMF_EFFC.TXT, trial and error and some invaluable hints by Zatzen)
  * The OpenMPT source code is released under the BSD license. Read LICENSE for more details.
  */
@@ -15,17 +15,16 @@
 
 #include "stdafx.h"
 #include "Loaders.h"
-#include "ChunkReader.h"
-#include <stdexcept>
+#include "BitReader.h"
 
 OPENMPT_NAMESPACE_BEGIN
 
 // DMF header
 struct DMFFileHeader
 {
-	char   signature[4];	// "DDMF"
-	uint8  version;			// 1 - 7 are beta versions, 8 is the official thing, 10 is xtracker32
-	char   tracker[8];		// "XTRACKER"
+	char   signature[4];  // "DDMF"
+	uint8  version;       // 1 - 7 are beta versions, 8 is the official thing, 10 is xtracker32
+	char   tracker[8];    // "XTRACKER"
 	char   songname[30];
 	char   composer[20];
 	uint8  creationDay;
@@ -40,14 +39,14 @@ struct DMFChunk
 	// 32-Bit chunk identifiers
 	enum ChunkIdentifiers
 	{
-		idCMSG	= MAGIC4LE('C','M','S','G'),	// Song message
-		idSEQU	= MAGIC4LE('S','E','Q','U'),	// Order list
-		idPATT	= MAGIC4LE('P','A','T','T'),	// Patterns
-		idSMPI	= MAGIC4LE('S','M','P','I'),	// Sample headers
-		idSMPD	= MAGIC4LE('S','M','P','D'),	// Sample data
-		idSMPJ	= MAGIC4LE('S','M','P','J'),	// Sample jump table (XTracker 32 only)
-		idENDE	= MAGIC4LE('E','N','D','E'),	// Last four bytes of DMF file
-		idSETT	= MAGIC4LE('S','E','T','T'),	// Probably contains GUI settings
+		idCMSG = MagicLE("CMSG"),  // Song message
+		idSEQU = MagicLE("SEQU"),  // Order list
+		idPATT = MagicLE("PATT"),  // Patterns
+		idSMPI = MagicLE("SMPI"),  // Sample headers
+		idSMPD = MagicLE("SMPD"),  // Sample data
+		idSMPJ = MagicLE("SMPJ"),  // Sample jump table (XTracker 32 only)
+		idENDE = MagicLE("ENDE"),  // Last four bytes of DMF file
+		idSETT = MagicLE("SETT"),  // Probably contains GUI settings
 	};
 
 	uint32le id;
@@ -66,21 +65,11 @@ struct DMFChunk
 
 MPT_BINARY_STRUCT(DMFChunk, 8)
 
-// Order list
-struct DMFSequence
-{
-	uint16le loopStart;
-	uint16le loopEnd;
-	// order list follows here ...
-};
-
-MPT_BINARY_STRUCT(DMFSequence, 4)
-
 // Pattern header (global)
 struct DMFPatterns
 {
-	uint16le numPatterns;	// 1..1024 patterns
-	uint8le  numTracks;		// 1..32 channels
+	uint16le numPatterns;  // 1..1024 patterns
+	uint8le  numTracks;    // 1..32 channels
 };
 
 MPT_BINARY_STRUCT(DMFPatterns, 3)
@@ -88,8 +77,8 @@ MPT_BINARY_STRUCT(DMFPatterns, 3)
 // Pattern header (for each pattern)
 struct DMFPatternHeader
 {
-	uint8le  numTracks;	// 1..32 channels
-	uint8le  beat;		// [hi|lo] -> hi = rows per beat, lo = reserved
+	uint8le  numTracks;  // 1..32 channels
+	uint8le  beat;       // [hi|lo] -> hi = rows per beat, lo = reserved
 	uint16le numRows;
 	uint32le patternLength;
 	// patttern data follows here ...
@@ -103,20 +92,20 @@ struct DMFSampleHeader
 	enum SampleFlags
 	{
 		// Sample flags
-		smpLoop		= 0x01,
-		smp16Bit	= 0x02,
+		smpLoop     = 0x01,
+		smp16Bit    = 0x02,
 		smpCompMask = 0x0C,
-		smpComp1	= 0x04,	// Compression type 1
-		smpComp2	= 0x08,	// Compression type 2 (unused)
-		smpComp3	= 0x0C,	// Compression type 3 (ditto)
-		smpLibrary	= 0x80,	// Sample is stored in a library
+		smpComp1    = 0x04,  // Compression type 1
+		smpComp2    = 0x08,  // Compression type 2 (unused)
+		smpComp3    = 0x0C,  // Compression type 3 (ditto)
+		smpLibrary  = 0x80,  // Sample is stored in a library
 	};
 
 	uint32le length;
 	uint32le loopStart;
 	uint32le loopEnd;
-	uint16le c3freq;		// 1000..45000hz
-	uint8le  volume;		// 0 = ignore
+	uint16le c3freq;  // 1000..45000hz
+	uint8le  volume;  // 0 = ignore
 	uint8le  flags;
 
 	// Convert an DMFSampleHeader to OpenMPT's internal sample representation.
@@ -151,53 +140,30 @@ struct DMFSampleHeader
 
 MPT_BINARY_STRUCT(DMFSampleHeader, 16)
 
-// Sample header tail (between head and tail, there might be the library name of the sample, depending on the DMF version)
-struct DMFSampleHeaderTail
-{
-	uint16le filler;
-	uint32le crc32;
-};
-
-MPT_BINARY_STRUCT(DMFSampleHeaderTail, 6)
-
 
 // Pattern translation memory
 struct DMFPatternSettings
 {
 	struct ChannelState
 	{
-		ModCommand::NOTE noteBuffer;	// Note buffer
-		ModCommand::NOTE lastNote;		// Last played note on channel
-		uint8 vibratoType;				// Last used vibrato type on channel
-		uint8 tremoloType;				// Last used tremolo type on channel
-		uint8 highOffset;				// Last used high offset on channel
-		bool playDir;					// Sample play direction... false = forward (default)
-
-		ChannelState()
-		{
-			noteBuffer = lastNote = NOTE_NONE;
-			vibratoType = 8;
-			tremoloType = 4;
-			highOffset = 6;
-			playDir = false;
-		}
+		ModCommand::NOTE noteBuffer = NOTE_NONE; // Note buffer
+		ModCommand::NOTE lastNote = NOTE_NONE;   // Last played note on channel
+		uint8 vibratoType = 8;                   // Last used vibrato type on channel
+		uint8 tremoloType = 4;                   // Last used tremolo type on channel
+		uint8 highOffset = 6;                    // Last used high offset on channel
+		bool playDir = false;                    // Sample play direction... false = forward (default)
 	};
 
-	std::vector<ChannelState> channels;		// Memory for each channel's state
-	bool realBPMmode;						// true = BPM mode
-	uint8 beat;								// Rows per beat
-	uint8 tempoTicks;						// Tick mode param
-	uint8 tempoBPM;							// BPM mode param
-	uint8 internalTicks;					// Ticks per row in final pattern
+	std::vector<ChannelState> channels; // Memory for each channel's state
+	bool realBPMmode = false;           // true = BPM mode
+	uint8 beat = 0;                     // Rows per beat
+	uint8 tempoTicks = 32;              // Tick mode param
+	uint8 tempoBPM = 120;               // BPM mode param
+	uint8 internalTicks = 6;            // Ticks per row in final pattern
 
-	DMFPatternSettings(CHANNELINDEX numChannels) : channels(numChannels)
-	{
-		realBPMmode = false;
-		beat = 0;
-		tempoTicks = 32;
-		tempoBPM = 120;
-		internalTicks = 6;
-	}
+	DMFPatternSettings(CHANNELINDEX numChannels)
+	    : channels(numChannels)
+	{ }
 };
 
 
@@ -209,17 +175,17 @@ static uint8 DMFporta2MPT(uint8 val, const uint8 internalTicks, const bool hasFi
 	else if((val <= 0x0F && hasFine) || internalTicks < 2)
 		return (val | 0xF0);
 	else
-		return std::max<uint8>(1, (val / (internalTicks - 1)));	// no porta on first tick!
+		return std::max(uint8(1), static_cast<uint8>((val / (internalTicks - 1))));	// no porta on first tick!
 }
 
 
 // Convert portamento / volume slide value (not very accurate due to X-Tracker's higher granularity, to say the least)
 static uint8 DMFslide2MPT(uint8 val, const uint8 internalTicks, const bool up)
 {
-	val = std::max<uint8>(1, val / 4);
+	val = std::max(uint8(1), static_cast<uint8>(val / 4));
 	const bool isFine = (val < 0x0F) || (internalTicks < 2);
 	if(!isFine)
-		val = std::max<uint8>(1, (val + internalTicks - 2) / (internalTicks - 1));	// no slides on first tick! "+ internalTicks - 2" for rounding precision
+		val = std::max(uint8(1), static_cast<uint8>((val + internalTicks - 2) / (internalTicks - 1)));	// no slides on first tick! "+ internalTicks - 2" for rounding precision
 
 	if(up)
 		return (isFine ? 0x0F : 0x00) | (val << 4);
@@ -254,19 +220,17 @@ static uint8 DMFvibrato2MPT(uint8 val, const uint8 internalTicks)
 {
 	// MPT: 1 vibrato period == 64 ticks... we have internalTicks ticks per row.
 	// X-Tracker: Period length specified in rows!
-	const int periodInTicks = MAX(1, (val >> 4)) * internalTicks;
-	const uint8 matchingPeriod = (uint8)Clamp((128 / periodInTicks), 1, 15);
-	return (matchingPeriod << 4) | MAX(1, (val & 0x0F));
+	const int periodInTicks = std::max(1, (val >> 4)) * internalTicks;
+	const uint8 matchingPeriod = static_cast<uint8>(Clamp((128 / periodInTicks), 1, 15));
+	return (matchingPeriod << 4) | std::max(uint8(1), static_cast<uint8>(val & 0x0F));
 }
 
 
 // Try using effect memory (zero paramer) to give the effect swapper some optimization hints.
-static void ApplyEffectMemory(const ModCommand *m, ROWINDEX row, CHANNELINDEX numChannels, uint8 effect, uint8 &param)
+static void ApplyEffectMemory(const ModCommand *m, ROWINDEX row, CHANNELINDEX numChannels, EffectCommand effect, uint8 &param)
 {
 	if(effect == CMD_NONE || param == 0)
-	{
 		return;
-	}
 
 	const bool isTonePortaEffect = (effect == CMD_PORTAMENTOUP || effect == CMD_PORTAMENTODOWN || effect == CMD_TONEPORTAMENTO);
 	const bool isVolSlideEffect = (effect == CMD_VOLUMESLIDE || effect == CMD_TONEPORTAVOL || effect == CMD_VIBRATOVOL);
@@ -323,28 +287,39 @@ static void ApplyEffectMemory(const ModCommand *m, ROWINDEX row, CHANNELINDEX nu
 }
 
 
-static PATTERNINDEX ConvertDMFPattern(FileReader &file, DMFPatternSettings &settings, CSoundFile &sndFile)
+static PATTERNINDEX ConvertDMFPattern(FileReader &file, const uint8 fileVersion, DMFPatternSettings &settings, CSoundFile &sndFile)
 {
 	// Pattern flags
 	enum PatternFlags
 	{
 		// Global Track
-		patGlobPack	= 0x80,	// Pack information for global track follows
-		patGlobMask	= 0x3F,	// Mask for global effects
+		patGlobPack = 0x80,  // Pack information for global track follows
+		patGlobMask = 0x3F,  // Mask for global effects
 		// Note tracks
-		patCounter	= 0x80,	// Pack information for current channel follows
-		patInstr	= 0x40,	// Instrument number present
-		patNote		= 0x20,	// Note present
-		patVolume	= 0x10,	// Volume present
-		patInsEff	= 0x08,	// Instrument effect present
-		patNoteEff	= 0x04,	// Note effect present
-		patVolEff	= 0x02,	// Volume effect stored
+		patCounter = 0x80,  // Pack information for current channel follows
+		patInstr   = 0x40,  // Instrument number present
+		patNote    = 0x20,  // Note present
+		patVolume  = 0x10,  // Volume present
+		patInsEff  = 0x08,  // Instrument effect present
+		patNoteEff = 0x04,  // Note effect present
+		patVolEff  = 0x02,  // Volume effect stored
 	};
 
 	file.Rewind();
 	
 	DMFPatternHeader patHead;
-	file.ReadStruct(patHead);
+	if(fileVersion < 3)
+	{
+		patHead.numTracks = file.ReadUint8();
+		file.Skip(2);  // not sure what this is, later X-Tracker versions just skip over it
+		patHead.numRows = file.ReadUint16LE();
+		patHead.patternLength = file.ReadUint32LE();
+	} else
+	{
+		file.ReadStruct(patHead);
+	}
+	if(fileVersion < 6)
+		patHead.beat = 0;
 
 	const ROWINDEX numRows = Clamp(ROWINDEX(patHead.numRows), ROWINDEX(1), MAX_PATTERN_ROWS);
 	const PATTERNINDEX pat = sndFile.Patterns.InsertAny(numRows);
@@ -353,8 +328,8 @@ static PATTERNINDEX ConvertDMFPattern(FileReader &file, DMFPatternSettings &sett
 		return pat;
 	}
 
-	PatternRow m = sndFile.Patterns[pat].GetRow(0);
-	const CHANNELINDEX numChannels = std::min<CHANNELINDEX>(sndFile.GetNumChannels() - 1, patHead.numTracks);
+	ModCommand *m = sndFile.Patterns[pat].GetpModCommand(0, 0);
+	const CHANNELINDEX numChannels = std::min(static_cast<CHANNELINDEX>(sndFile.GetNumChannels() - 1), static_cast<CHANNELINDEX>(patHead.numTracks));
 
 	// When breaking to a pattern with less channels that the previous pattern,
 	// all voices in the now unused channels are killed:
@@ -393,17 +368,17 @@ static PATTERNINDEX ConvertDMFPattern(FileReader &file, DMFPatternSettings &sett
 
 			switch(globalInfo)
 			{
-			case 1:		// Set Tick Frame Speed
+			case 1:  // Set Tick Frame Speed
 				settings.realBPMmode = false;
-				settings.tempoTicks = std::max(uint8(1), globalData);	// Tempo in 1/4 rows per second
-				settings.tempoBPM = 0;									// Automatically updated by X-Tracker
+				settings.tempoTicks = std::max(uint8(1), globalData);  // Tempo in 1/4 rows per second
+				settings.tempoBPM = 0;                                 // Automatically updated by X-Tracker
 				tempoChange = true;
 				break;
-			case 2:		// Set BPM Speed (real BPM mode)
-				if(globalData)	// DATA = 0 doesn't do anything
+			case 2:             // Set BPM Speed (real BPM mode)
+				if(globalData)  // DATA = 0 doesn't do anything
 				{
 					settings.realBPMmode = true;
-					settings.tempoBPM = globalData;		// Tempo in real BPM (depends on rows per beat)
+					settings.tempoBPM = globalData;  // Tempo in real BPM (depends on rows per beat)
 					if(settings.beat != 0)
 					{
 						settings.tempoTicks = (globalData * settings.beat * 15);	// Automatically updated by X-Tracker
@@ -411,7 +386,7 @@ static PATTERNINDEX ConvertDMFPattern(FileReader &file, DMFPatternSettings &sett
 					tempoChange = true;
 				}
 				break;
-			case 3:		// Set Beat
+			case 3:  // Set Beat
 				settings.beat = (globalData >> 4);
 				if(settings.beat != 0)
 				{
@@ -423,12 +398,12 @@ static PATTERNINDEX ConvertDMFPattern(FileReader &file, DMFPatternSettings &sett
 					settings.realBPMmode = false;
 				}
 				break;
-			case 4:		// Tick Delay
+			case 4:  // Tick Delay
 				writeDelay = globalData;
 				break;
-			case 5:		// Set External Flag
+			case 5:  // Set External Flag
 				break;
-			case 6:		// Slide Speed Up
+			case 6:  // Slide Speed Up
 				if(globalData > 0)
 				{
 					uint8 &tempoData = (settings.realBPMmode) ? settings.tempoBPM : settings.tempoTicks;
@@ -442,7 +417,7 @@ static PATTERNINDEX ConvertDMFPattern(FileReader &file, DMFPatternSettings &sett
 					tempoChange = true;
 				}
 				break;
-			case 7:		// Slide Speed Down
+			case 7:  // Slide Speed Down
 				if(globalData > 0)
 				{
 					uint8 &tempoData = (settings.realBPMmode) ? settings.tempoBPM : settings.tempoTicks;
@@ -477,7 +452,7 @@ static PATTERNINDEX ConvertDMFPattern(FileReader &file, DMFPatternSettings &sett
 				// => Tempo = 60 * Rows per Second * Speed / 24
 				// For some reason, using settings.tempoTicks + 1 gives more accurate results than just settings.tempoTicks... (same problem in the old libmodplug DMF loader)
 				// Original unoptimized formula:
-				//const int tickspeed = (tempoRealBPMmode) ? MAX(1, (tempoData * beat * 4) / 60) : tempoData;
+				//const int tickspeed = (tempoRealBPMmode) ? std::max(1, (tempoData * beat * 4) / 60) : tempoData;
 				const int tickspeed = (settings.realBPMmode) ? std::max(1, settings.tempoBPM * settings.beat * 2) : ((settings.tempoTicks + 1) * 30);
 				// Try to find matching speed - try higher speeds first, so that effects like arpeggio and tremor work better.
 				for(speed = 255; speed >= 1; speed--)
@@ -496,7 +471,7 @@ static PATTERNINDEX ConvertDMFPattern(FileReader &file, DMFPatternSettings &sett
 			}
 		}
 
-		m = sndFile.Patterns[pat].GetpModCommand(row, 1);	// Reserve first channel for global effects
+		m = sndFile.Patterns[pat].GetpModCommand(row, 1);  // Reserve first channel for global effects
 
 		for(CHANNELINDEX chn = 1; chn <= numChannels; chn++, m++)
 		{
@@ -513,7 +488,7 @@ static PATTERNINDEX ConvertDMFPattern(FileReader &file, DMFPatternSettings &sett
 
 				////////////////////////////////////////////////////////////////
 				// 0x40: Instrument
-				bool slideNote = true;		// If there is no instrument number next to a note, the note is not retriggered!
+				bool slideNote = true;  // If there is no instrument number next to a note, the note is not retriggered!
 				if((channelInfo & patInstr) != 0)
 				{
 					m->instr = file.ReadUint8();
@@ -556,7 +531,7 @@ static PATTERNINDEX ConvertDMFPattern(FileReader &file, DMFPatternSettings &sett
 					settings.channels[chn].playDir = false;
 				}
 
-				uint8 effect1 = CMD_NONE, effect2 = CMD_NONE, effect3 = CMD_NONE;
+				EffectCommand effect1 = CMD_NONE, effect2 = CMD_NONE, effect3 = CMD_NONE;
 				uint8 effectParam1 = 0, effectParam2 = 0, effectParam3 = 0;
 				bool useMem2 = false, useMem3 = false;	// Effect can use memory if necessary
 
@@ -565,32 +540,32 @@ static PATTERNINDEX ConvertDMFPattern(FileReader &file, DMFPatternSettings &sett
 				if((channelInfo & patVolume) != 0)
 				{
 					m->volcmd = VOLCMD_VOLUME;
-					m->vol = (file.ReadUint8() + 2) / 4;	// Should be + 3 instead of + 2, but volume 1 is silent in X-Tracker.
+					m->vol = (file.ReadUint8() + 2) / 4;  // Should be + 3 instead of + 2, but volume 1 is silent in X-Tracker.
 				}
 
 				////////////////////////////////////////////////////////////////
 				// 0x08: Instrument effect
 				if((channelInfo & patInsEff) != 0)
 				{
-					effect1 = file.ReadUint8();
+					const uint8 command = file.ReadUint8();
 					effectParam1 = file.ReadUint8();
 
-					switch(effect1)
+					switch(command)
 					{
-					case 1:		// Stop Sample
+					case 1:  // Stop Sample
 						m->note = NOTE_NOTECUT;
 						effect1 = CMD_NONE;
 						break;
-					case 2:		// Stop Sample Loop
+					case 2:  // Stop Sample Loop
 						m->note = NOTE_KEYOFF;
 						effect1 = CMD_NONE;
 						break;
-					case 3:		// Instrument Volume Override (aka "Restart")
+					case 3:  // Instrument Volume Override (aka "Restart")
 						m->note = settings.channels[chn].lastNote;
 						settings.channels[chn].playDir = false;
 						effect1 = CMD_NONE;
 						break;
-					case 4:		// Sample Delay
+					case 4:  // Sample Delay
 						effectParam1 = DMFdelay2MPT(effectParam1, settings.internalTicks);
 						if(effectParam1)
 						{
@@ -606,21 +581,21 @@ static PATTERNINDEX ConvertDMFPattern(FileReader &file, DMFPatternSettings &sett
 							settings.channels[chn].playDir = false;
 						}
 						break;
-					case 5:		// Tremolo Retrig Sample (who invented those stupid effect names?)
-						effectParam1 = MAX(1, DMFdelay2MPT(effectParam1, settings.internalTicks));
+					case 5:  // Tremolo Retrig Sample (who invented those stupid effect names?)
+						effectParam1 = std::max(uint8(1), DMFdelay2MPT(effectParam1, settings.internalTicks));
 						effect1 = CMD_RETRIG;
 						settings.channels[chn].playDir = false;
 						break;
-					case 6:		// Offset
-					case 7:		// Offset + 64k
-					case 8:		// Offset + 128k
-					case 9:		// Offset + 192k
+					case 6:  // Offset
+					case 7:  // Offset + 64k
+					case 8:  // Offset + 128k
+					case 9:  // Offset + 192k
 						// Put high offset on previous row
-						if(row > 0 && effect1 != settings.channels[chn].highOffset)
+						if(row > 0 && command != settings.channels[chn].highOffset)
 						{
-							if(sndFile.Patterns[pat].WriteEffect(EffectWriter(CMD_S3MCMDEX, (0xA0 | (effect1 - 6))).Row(row - 1).Channel(chn).RetryPreviousRow()))
+							if(sndFile.Patterns[pat].WriteEffect(EffectWriter(CMD_S3MCMDEX, (0xA0 | (command - 6))).Row(row - 1).Channel(chn).RetryPreviousRow()))
 							{
-								settings.channels[chn].highOffset = effect1;
+								settings.channels[chn].highOffset = command;
 							}
 						}
 						effect1 = CMD_OFFSET;
@@ -649,26 +624,21 @@ static PATTERNINDEX ConvertDMFPattern(FileReader &file, DMFPatternSettings &sett
 				// 0x04: Note effect
 				if((channelInfo & patNoteEff) != 0)
 				{
-					effect2 = file.ReadUint8();
+					const uint8 command = file.ReadUint8();
 					effectParam2 = file.ReadUint8();
 
-					switch(effect2)
+					switch(command)
 					{
 					case 1:  // Note Finetune (1/16th of a semitone signed 8-bit value, not 1/128th as the interface claims)
-						effect2 = (effectParam2 < 128) ? CMD_PORTAMENTOUP : CMD_PORTAMENTODOWN;
-						if(effectParam2 >= 128)
-							effectParam2 = ~effectParam2 + 1;
-						if(effectParam2 >= 16 && m->IsNote())
 						{
-							if(effect2 == CMD_PORTAMENTOUP)
-								m->note = static_cast<ModCommand::NOTE>(std::min(m->note + effectParam2 / 16, static_cast<int>(NOTE_MAX)));
-							else
-								m->note = static_cast<ModCommand::NOTE>(std::max(m->note - effectParam2 / 16, static_cast<int>(NOTE_MIN)));
-							effectParam2 %= 16u;
+							const auto fine = std::div(static_cast<int8>(effectParam2) * 8, 128);
+							if(m->IsNote())
+								m->note = static_cast<ModCommand::NOTE>(Clamp(m->note + fine.quot, NOTE_MIN, NOTE_MAX));
+							effect2 = CMD_FINETUNE;
+							effectParam2 = static_cast<uint8>(fine.rem) ^ 0x80;
 						}
-						effectParam2 = 0xF0 | std::min(uint8(0x0F), effectParam2);
 						break;
-					case 2:		// Note Delay (wtf is the difference to Sample Delay?)
+					case 2:  // Note Delay (wtf is the difference to Sample Delay?)
 						effectParam2 = DMFdelay2MPT(effectParam2, settings.internalTicks);
 						if(effectParam2)
 						{
@@ -680,17 +650,17 @@ static PATTERNINDEX ConvertDMFPattern(FileReader &file, DMFPatternSettings &sett
 						}
 						useMem2 = true;
 						break;
-					case 3:		// Arpeggio
+					case 3:  // Arpeggio
 						effect2 = CMD_ARPEGGIO;
 						useMem2 = true;
 						break;
-					case 4:		// Portamento Up
-					case 5:		// Portamento Down
+					case 4:  // Portamento Up
+					case 5:  // Portamento Down
 						effectParam2 = DMFporta2MPT(effectParam2, settings.internalTicks, true);
-						effect2 = (effect2 == 4) ? CMD_PORTAMENTOUP : CMD_PORTAMENTODOWN;
+						effect2 = (command == 4) ? CMD_PORTAMENTOUP : CMD_PORTAMENTODOWN;
 						useMem2 = true;
 						break;
-					case 6:		// Portamento to Note
+					case 6:  // Portamento to Note
 						if(m->note == NOTE_NONE)
 						{
 							m->note = settings.channels[chn].noteBuffer;
@@ -699,33 +669,33 @@ static PATTERNINDEX ConvertDMFPattern(FileReader &file, DMFPatternSettings &sett
 						effect2 = CMD_TONEPORTAMENTO;
 						useMem2 = true;
 						break;
-					case 7:		// Scratch to Note (neat! but we don't have such an effect...)
+					case 7:  // Scratch to Note (neat! but we don't have such an effect...)
 						m->note = static_cast<ModCommand::NOTE>(Clamp(effectParam2 + 25, NOTE_MIN, NOTE_MAX));
 						effect2 = CMD_TONEPORTAMENTO;
 						effectParam2 = 0xFF;
 						useMem2 = true;
 						break;
-					case 8:		// Vibrato Sine
-					case 9:		// Vibrato Triangle (ramp down should be close enough)
-					case 10:	// Vibrato Square
+					case 8:   // Vibrato Sine
+					case 9:   // Vibrato Triangle (ramp down should be close enough)
+					case 10:  // Vibrato Square
 						// Put vibrato type on previous row
-						if(row > 0 && effect2 != settings.channels[chn].vibratoType)
+						if(row > 0 && command != settings.channels[chn].vibratoType)
 						{
-							if(sndFile.Patterns[pat].WriteEffect(EffectWriter(CMD_S3MCMDEX, (0x30 | (effect2 - 8))).Row(row - 1).Channel(chn).RetryPreviousRow()))
+							if(sndFile.Patterns[pat].WriteEffect(EffectWriter(CMD_S3MCMDEX, (0x30 | (command - 8))).Row(row - 1).Channel(chn).RetryPreviousRow()))
 							{
-								settings.channels[chn].vibratoType = effect2;
+								settings.channels[chn].vibratoType = command;
 							}
 						}
 						effect2 = CMD_VIBRATO;
 						effectParam2 = DMFvibrato2MPT(effectParam2, settings.internalTicks);
 						useMem2 = true;
 						break;
-					case 11:	 // Note Tremolo
+					case 11:  // Note Tremolo
 						effectParam2 = DMFtremor2MPT(effectParam2, settings.internalTicks);
 						effect2 = CMD_TREMOR;
 						useMem2 = true;
 						break;
-					case 12:	// Note Cut
+					case 12:  // Note Cut
 						effectParam2 = DMFdelay2MPT(effectParam2, settings.internalTicks);
 						if(effectParam2)
 						{
@@ -748,47 +718,47 @@ static PATTERNINDEX ConvertDMFPattern(FileReader &file, DMFPatternSettings &sett
 				// 0x02: Volume effect
 				if((channelInfo & patVolEff) != 0)
 				{
-					effect3 = file.ReadUint8();
+					const uint8 command = file.ReadUint8();
 					effectParam3 = file.ReadUint8();
 
-					switch(effect3)
+					switch(command)
 					{
-					case 1:		// Volume Slide Up
-					case 2:		// Volume Slide Down
-						effectParam3 = DMFslide2MPT(effectParam3, settings.internalTicks, (effect3 == 1));
+					case 1:  // Volume Slide Up
+					case 2:  // Volume Slide Down
+						effectParam3 = DMFslide2MPT(effectParam3, settings.internalTicks, (command == 1));
 						effect3 = CMD_VOLUMESLIDE;
 						useMem3 = true;
 						break;
-					case 3:		// Volume Tremolo (actually this is Tremor)
+					case 3:  // Volume Tremolo (actually this is Tremor)
 						effectParam3 = DMFtremor2MPT(effectParam3, settings.internalTicks);
 						effect3 = CMD_TREMOR;
 						useMem3 = true;
 						break;
-					case 4:		// Tremolo Sine
-					case 5:		// Tremolo Triangle (ramp down should be close enough)
-					case 6:		// Tremolo Square
+					case 4:  // Tremolo Sine
+					case 5:  // Tremolo Triangle (ramp down should be close enough)
+					case 6:  // Tremolo Square
 						// Put tremolo type on previous row
-						if(row > 0 && effect3 != settings.channels[chn].tremoloType)
+						if(row > 0 && command != settings.channels[chn].tremoloType)
 						{
-							if(sndFile.Patterns[pat].WriteEffect(EffectWriter(CMD_S3MCMDEX, (0x40 | (effect3 - 4))).Row(row - 1).Channel(chn).RetryPreviousRow()))
+							if(sndFile.Patterns[pat].WriteEffect(EffectWriter(CMD_S3MCMDEX, (0x40 | (command - 4))).Row(row - 1).Channel(chn).RetryPreviousRow()))
 							{
-								settings.channels[chn].tremoloType = effect3;
+								settings.channels[chn].tremoloType = command;
 							}
 						}
 						effect3 = CMD_TREMOLO;
 						effectParam3 = DMFvibrato2MPT(effectParam3, settings.internalTicks);
 						useMem3 = true;
 						break;
-					case 7:		// Set Balance
+					case 7:  // Set Balance
 						effect3 = CMD_PANNING8;
 						break;
-					case 8:		// Slide Balance Left
-					case 9:		// Slide Balance Right
-						effectParam3 = DMFslide2MPT(effectParam3, settings.internalTicks, (effect3 == 8));
+					case 8:  // Slide Balance Left
+					case 9:  // Slide Balance Right
+						effectParam3 = DMFslide2MPT(effectParam3, settings.internalTicks, (command == 8));
 						effect3 = CMD_PANNINGSLIDE;
 						useMem3 = true;
 						break;
-					case 10:	// Balance Vibrato Left/Right (always sine modulated)
+					case 10:  // Balance Vibrato Left/Right (always sine modulated)
 						effect3 = CMD_PANBRELLO;
 						effectParam3 = DMFvibrato2MPT(effectParam3, settings.internalTicks);
 						useMem3 = true;
@@ -801,13 +771,9 @@ static PATTERNINDEX ConvertDMFPattern(FileReader &file, DMFPatternSettings &sett
 
 				// Let's see if we can help the effect swapper by reducing some effect parameters to "continue" parameters.
 				if(useMem2)
-				{
 					ApplyEffectMemory(m, row, sndFile.GetNumChannels(), effect2, effectParam2);
-				}
 				if(useMem3)
-				{
 					ApplyEffectMemory(m, row, sndFile.GetNumChannels(), effect3, effectParam3);
-				}
 
 				// I guess this is close enough to "not retriggering the note"
 				if(slideNote && m->IsNote())
@@ -838,29 +804,32 @@ static PATTERNINDEX ConvertDMFPattern(FileReader &file, DMFPatternSettings &sett
 					}
 				}
 
-				ModCommand::TwoRegularCommandsToMPT(effect2, effectParam2, effect3, effectParam3);
+				ModCommand combinedCmd;
+				combinedCmd.FillInTwoCommands(effect2, effectParam2, effect3, effectParam3);
 
-				if(m->volcmd == VOLCMD_NONE && effect2 != VOLCMD_NONE)
+				if(m->volcmd == VOLCMD_NONE && combinedCmd.volcmd != VOLCMD_NONE)
 				{
-					m->volcmd = effect2;
-					m->vol = effectParam2;
+					m->SetVolumeCommand(combinedCmd);
 				}
 				// Prefer instrument effects over any other effects
 				if(effect1 != CMD_NONE)
 				{
-					m->command = effect1;
-					m->param = effectParam1;
-				} else if(effect3 != CMD_NONE)
+					combinedCmd.FillInTwoCommands(combinedCmd.command, combinedCmd.param, effect1, effectParam1);
+					if(m->volcmd == VOLCMD_NONE && combinedCmd.volcmd != VOLCMD_NONE)
+					{
+						m->SetVolumeCommand(combinedCmd);
+					}
+					m->SetEffectCommand(combinedCmd);
+				} else if(combinedCmd.command != CMD_NONE)
 				{
-					m->command = effect3;
-					m->param = effectParam3;
+					m->SetEffectCommand(combinedCmd);
 				}
 
 			} else
 			{
 				channelCounter[chn]--;
 			}
-		}	// End for all channels
+		}  // End for all channels
 
 		// Now we can try to write tempo information.
 		if(tempoChange)
@@ -881,7 +850,7 @@ static PATTERNINDEX ConvertDMFPattern(FileReader &file, DMFPatternSettings &sett
 			sndFile.Patterns[pat].WriteEffect(EffectWriter(CMD_S3MCMDEX, 0x60u | Clamp(param, uint8(1), uint8(15))).Row(row).AllowMultiple());
 		}
 		writeDelay = 0;
-	}	// End for all rows
+	}  // End for all rows
 
 	return pat;
 }
@@ -933,34 +902,58 @@ bool CSoundFile::ReadDMF(FileReader &file, ModLoadingFlags loadFlags)
 	}
 
 	InitializeGlobals(MOD_TYPE_DMF);
-	mpt::String::Read<mpt::String::spacePadded>(m_songName, fileHeader.songname);
-	{
-		std::string artist;
-		mpt::String::Read<mpt::String::spacePadded>(artist, fileHeader.composer);
-		m_songArtist = mpt::ToUnicode(mpt::CharsetCP437, artist);
-	}
+
+	m_modFormat.formatName = MPT_UFORMAT("X-Tracker v{}")(fileHeader.version);
+	m_modFormat.type = U_("dmf");
+	m_modFormat.charset = mpt::Charset::CP437;
+
+	m_songName = mpt::String::ReadBuf(mpt::String::spacePadded, fileHeader.songname);
+	m_songArtist = mpt::ToUnicode(mpt::Charset::CP437, mpt::String::ReadBuf(mpt::String::spacePadded, fileHeader.composer));
 
 	FileHistory mptHistory;
-	MemsetZero(mptHistory);
-	mptHistory.loadDate.tm_mday = Clamp(fileHeader.creationDay, uint8(1), uint8(31));
-	mptHistory.loadDate.tm_mon = Clamp(fileHeader.creationMonth, uint8(1), uint8(12)) - 1;
-	mptHistory.loadDate.tm_year = fileHeader.creationYear;
+	mptHistory.loadDate.day = Clamp(fileHeader.creationDay, uint8(1), uint8(31));
+	mptHistory.loadDate.month = Clamp(fileHeader.creationMonth, uint8(1), uint8(12));
+	mptHistory.loadDate.year = 1900 + fileHeader.creationYear;
 	m_FileHistory.clear();
 	m_FileHistory.push_back(mptHistory);
 
-	// Go through all chunks now
-	ChunkReader chunkFile(file);
-	ChunkReader::ChunkList<DMFChunk> chunks = chunkFile.ReadChunks<DMFChunk>(1);
+	// Go through all chunks now... cannot use our standard IFF chunk reader here because early X-Tracker versions write some malformed chunk headers... fun code ahead!
+	ChunkReader::ChunkList<DMFChunk> chunks;
+	while(file.CanRead(sizeof(DMFChunk)))
+	{
+		DMFChunk chunkHeader;
+		file.Read(chunkHeader);
+		uint32 chunkLength = chunkHeader.length, chunkSkip = 0;
+		// When loop start was added to version 3, the chunk size was not updated...
+		if(fileHeader.version == 3 && chunkHeader.GetID() == DMFChunk::idSEQU)
+			chunkSkip = 2;
+		// ...and when the loop end was added to version 4, it was also note updated! Luckily they fixed it in version 5.
+		else if(fileHeader.version == 4 && chunkHeader.GetID() == DMFChunk::idSEQU)
+			chunkSkip = 4;
+		// Earlier X-Tracker versions also write a garbage length for the SMPD chunk if samples are compressed.
+		// I don't know when exactly this stopped, but I have no version 5-7 files to check (and no X-Tracker version that writes those versions).
+		// Since this is practically always the last chunk in the file, the following code is safe for those versions, though.
+		else if(fileHeader.version < 8 && chunkHeader.GetID() == DMFChunk::idSMPD)
+			chunkLength = mpt::saturate_cast<uint32>(file.BytesLeft());
+		chunks.chunks.push_back(ChunkReader::Item<DMFChunk>{chunkHeader, file.ReadChunk(chunkLength)});
+		file.Skip(chunkSkip);
+	}
 	FileReader chunk;
 
 	// Read order list
-	DMFSequence seqHeader;
 	chunk = chunks.GetChunk(DMFChunk::idSEQU);
-	if(!chunk.ReadStruct(seqHeader))
-	{
-		return false;
-	}
-	ReadOrderFromFile<uint16le>(Order(), chunk, (chunk.GetLength() - sizeof(DMFSequence)) / 2);
+	ORDERINDEX seqLoopStart = 0, seqLoopEnd = ORDERINDEX_MAX;
+	if(fileHeader.version >= 3)
+		seqLoopStart = chunk.ReadUint16LE();
+	if(fileHeader.version >= 4)
+		seqLoopEnd = chunk.ReadUint16LE();
+	// HIPOMATK.DMF has a loop end of 0, other v4 files have proper loop ends. Later X-Tracker versions import it as-is but it cannot be intentional.
+	// We just assume that this feature might have been buggy in early v4 versions and ignore the loop end in that case.
+	if(fileHeader.version == 4 && seqLoopEnd == 0)
+		seqLoopEnd = ORDERINDEX_MAX;
+	ReadOrderFromFile<uint16le>(Order(), chunk, chunk.BytesLeft() / 2);
+	LimitMax(seqLoopStart, Order().GetLastIndex());
+	LimitMax(seqLoopEnd, Order().GetLastIndex());
 
 	// Read patterns
 	chunk = chunks.GetChunk(DMFChunk::idPATT);
@@ -970,37 +963,35 @@ bool CSoundFile::ReadDMF(FileReader &file, ModLoadingFlags loadFlags)
 		chunk.ReadStruct(patHeader);
 		m_nChannels = Clamp<uint8, uint8>(patHeader.numTracks, 1, 32) + 1;	// + 1 for global track (used for tempo stuff)
 
-		std::vector<FileReader> patternChunks;
-		patternChunks.reserve(patHeader.numPatterns);
-
 		// First, find out where all of our patterns are...
-		for(PATTERNINDEX pat = 0; pat < patHeader.numPatterns; pat++)
+		std::vector<FileReader> patternChunks(patHeader.numPatterns);
+		for(auto &patternChunk : patternChunks)
 		{
-			DMFPatternHeader header;
-			chunk.ReadStruct(header);
-			chunk.SkipBack(sizeof(header));
-			patternChunks.push_back(chunk.ReadChunk(sizeof(header) + header.patternLength));
+			const uint8 headerSize = fileHeader.version < 3 ? 9 : 8;
+			chunk.Skip(headerSize - sizeof(uint32le));
+			const uint32 patLength = chunk.ReadUint32LE();
+			chunk.SkipBack(headerSize);
+			patternChunk = chunk.ReadChunk(headerSize + patLength);
 		}
 
 		// Now go through the order list and load them.
 		DMFPatternSettings settings(GetNumChannels());
 
 		Patterns.ResizeArray(Order().GetLength());
-		for(ORDERINDEX ord = 0; ord < Order().GetLength(); ord++)
+		for(PATTERNINDEX &pat : Order())
 		{
 			// Create one pattern for each order item, as the same pattern can be played with different settings
-			PATTERNINDEX pat = Order()[ord];
 			if(pat < patternChunks.size())
 			{
-				pat = ConvertDMFPattern(patternChunks[pat], settings, *this);
-				Order()[ord] = pat;
-				// Loop end?
-				if(pat != PATTERNINDEX_INVALID && ord == seqHeader.loopEnd && (seqHeader.loopStart > 0 || ord < Order().GetLastIndex()))
-				{
-					Patterns[pat].WriteEffect(EffectWriter(CMD_POSITIONJUMP, static_cast<ModCommand::PARAM>(seqHeader.loopStart)).Row(Patterns[pat].GetNumRows() - 1).RetryPreviousRow());
-				}
+				pat = ConvertDMFPattern(patternChunks[pat], fileHeader.version, settings, *this);
 			}
 		}
+		// Write loop end if necessary
+		if(Order().IsValidPat(seqLoopEnd) && (seqLoopStart > 0 || seqLoopEnd < Order().GetLastIndex()))
+		{
+			PATTERNINDEX pat = Order()[seqLoopEnd];
+			Patterns[pat].WriteEffect(EffectWriter(CMD_POSITIONJUMP, static_cast<ModCommand::PARAM>(seqLoopStart)).Row(Patterns[pat].GetNumRows() - 1).RetryPreviousRow());
+		}
 	}
 
 	// Read song message
@@ -1011,7 +1002,7 @@ bool CSoundFile::ReadDMF(FileReader &file, ModLoadingFlags loadFlags)
 		// The skipped byte seems to always be 0.
 		// This also matches how XT 1.03 itself displays the song message.
 		chunk.Skip(1);
-		m_songMessage.ReadFixedLineLength(chunk, chunk.GetLength() - 1, 40, 0);
+		m_songMessage.ReadFixedLineLength(chunk, chunk.BytesLeft(), 40, 0);
 	}
 	
 	// Read sample headers + data
@@ -1021,20 +1012,19 @@ bool CSoundFile::ReadDMF(FileReader &file, ModLoadingFlags loadFlags)
 
 	for(SAMPLEINDEX smp = 1; smp <= GetNumSamples(); smp++)
 	{
-		chunk.ReadSizedString<uint8le, mpt::String::spacePadded>(m_szNames[smp]);
+		const uint8 nameLength = (fileHeader.version < 2) ? 30 : chunk.ReadUint8();
+		chunk.ReadString<mpt::String::spacePadded>(m_szNames[smp], nameLength);
 		DMFSampleHeader sampleHeader;
 		ModSample &sample = Samples[smp];
 		chunk.ReadStruct(sampleHeader);
 		sampleHeader.ConvertToMPT(sample);
 
+		// Read library name in version 8 files
 		if(fileHeader.version >= 8)
-		{
-			// Read library name in version 8 files
 			chunk.ReadString<mpt::String::spacePadded>(sample.filename, 8);
-		}
 
-		// We don't care for the checksum of the sample data...
-		chunk.Skip(sizeof(DMFSampleHeaderTail));
+		// Filler + CRC
+		chunk.Skip(fileHeader.version > 1 ? 6 : 2);
 
 		// Now read the sample data from the data chunk
 		FileReader sampleData = sampleDataChunk.ReadChunk(sampleDataChunk.ReadUint32LE());
@@ -1050,11 +1040,12 @@ bool CSoundFile::ReadDMF(FileReader &file, ModLoadingFlags loadFlags)
 	}
 
 	InitializeChannels();
-	m_SongFlags = SONG_LINEARSLIDES | SONG_ITCOMPATGXX;	// this will be converted to IT format by MPT. SONG_ITOLDEFFECTS is not set because of tremor and vibrato.
+	m_SongFlags = SONG_LINEARSLIDES | SONG_ITCOMPATGXX;  // this will be converted to IT format by MPT. SONG_ITOLDEFFECTS is not set because of tremor and vibrato.
 	m_nDefaultSpeed = 6;
 	m_nDefaultTempo.Set(120);
 	m_nDefaultGlobalVolume = 256;
 	m_nSamplePreAmp = m_nVSTiVolume = 48;
+	m_playBehaviour.set(kApplyOffsetWithoutNote);
 
 	return true;
 }
@@ -1071,53 +1062,31 @@ struct DMFHNode
 
 struct DMFHTree
 {
-	const uint8 *ibuf, *ibufmax;
-	uint32 bitbuf;
-	int bitnum;
-	int lastnode, nodecount;
-	DMFHNode nodes[256];
+	BitReader file;
+	int lastnode = 0, nodecount = 0;
+	DMFHNode nodes[256]{};
 
-	// DMF Huffman ReadBits
-	uint8 DMFReadBits(int nbits)
+	DMFHTree(FileReader &file)
+	    : file(file)
 	{
-		if(bitnum < nbits)
-		{
-			if(ibuf < ibufmax)
-			{
-				bitbuf |= (((uint32)(*ibuf++)) << bitnum);
-				bitnum += 8;
-			} else
-			{
-				throw std::range_error("Truncated DMF sample block");
-			}
-		}
-
-		uint8 v = static_cast<uint8>(bitbuf & ((1 << nbits) - 1));
-		bitbuf >>= nbits;
-		bitnum -= nbits;
-		return v;
 	}
-
-
+	
 	//
 	// tree: [8-bit value][12-bit index][12-bit index] = 32-bit
 	//
 
 	void DMFNewNode()
 	{
-		uint8 isleft, isright;
-		int actnode;
-
-		actnode = nodecount;
+		int actnode = nodecount;
 		if(actnode > 255) return;
-		nodes[actnode].value = DMFReadBits(7);
-		isleft = DMFReadBits(1);
-		isright = DMFReadBits(1);
+		nodes[actnode].value = static_cast<uint8>(file.ReadBits(7));
+		bool isLeft = file.ReadBits(1) != 0;
+		bool isRight = file.ReadBits(1) != 0;
 		actnode = lastnode;
 		if(actnode > 255) return;
 		nodecount++;
 		lastnode = nodecount;
-		if(isleft)
+		if(isLeft)
 		{
 			nodes[actnode].left = (int16)lastnode;
 			DMFNewNode();
@@ -1126,7 +1095,7 @@ struct DMFHTree
 			nodes[actnode].left = -1;
 		}
 		lastnode = nodecount;
-		if(isright)
+		if(isRight)
 		{
 			nodes[actnode].right = (int16)lastnode;
 			DMFNewNode();
@@ -1138,27 +1107,23 @@ struct DMFHTree
 };
 
 
-uintptr_t DMFUnpack(uint8 *psample, const uint8 *ibuf, const uint8 *ibufmax, uint32 maxlen)
+uintptr_t DMFUnpack(FileReader &file, uint8 *psample, uint32 maxlen)
 {
-	DMFHTree tree;
-
-	MemsetZero(tree);
-	tree.ibuf = ibuf;
-	tree.ibufmax = ibufmax;
-	tree.DMFNewNode();
+	DMFHTree tree(file);
 	uint8 value = 0, delta = 0;
 
 	try
 	{
+		tree.DMFNewNode();
 		if(tree.nodes[0].left < 0 || tree.nodes[0].right < 0)
-			return tree.ibuf - ibuf;
+			return tree.file.GetPosition();
 		for(uint32 i = 0; i < maxlen; i++)
 		{
 			int actnode = 0;
-			uint8 sign = tree.DMFReadBits(1);
+			bool sign = tree.file.ReadBits(1) != 0;
 			do
 			{
-				if(tree.DMFReadBits(1))
+				if(tree.file.ReadBits(1))
 					actnode = tree.nodes[actnode].right;
 				else
 					actnode = tree.nodes[actnode].left;
@@ -1169,11 +1134,11 @@ uintptr_t DMFUnpack(uint8 *psample, const uint8 *ibuf, const uint8 *ibufmax, uin
 			value += delta;
 			psample[i] = value;
 		}
-	} catch(const std::range_error &)
+	} catch(const BitReader::eof &)
 	{
 		//AddToLog(LogWarning, "Truncated DMF sample block");
 	}
-	return tree.ibuf - ibuf;
+	return tree.file.GetPosition();
 }
 
 
