@@ -12,11 +12,20 @@
 #include "Sndfile.h"
 #ifdef MODPLUG_TRACKER
 #include "../mptrack/TrackerSettings.h"
+#include "mpt/io_file/inputfile.hpp"
+#include "mpt/io_file_read/inputfile_filecursor.hpp"
+#include "../common/mptFileIO.h"
 #endif // MODPLUG_TRACKER
 #ifndef MODPLUG_NO_FILESAVE
+#include "mpt/io_file/outputfile.hpp"
 #include "../common/mptFileIO.h"
+#ifdef MODPLUG_TRACKER
+#include "mpt/fs/fs.hpp"
+#endif // MODPLUG_TRACKER
 #endif // !MODPLUG_NO_FILESAVE
 #include "modsmp_ctrl.h"
+#include "mpt/base/numbers.hpp"
+#include "mpt/parse/parse.hpp"
 
 #include <functional>
 
@@ -51,9 +60,9 @@ struct SFZControl
 		if(key == "default_path")
 			defaultPath = value;
 		else if(key == "octave_offset")
-			octaveOffset = ConvertStrTo<int8>(value);
+			octaveOffset = mpt::parse<int8>(value);
 		else if(key == "note_offset")
-			noteOffset = ConvertStrTo<int8>(value);
+			noteOffset = mpt::parse<int8>(value);
 	}
 };
 
@@ -71,7 +80,7 @@ struct SFZFlexEG
 	void Parse(std::string_view key, const std::string &value)
 	{
 		key = key.substr(key.find('_') + 1);
-		const double v = ConvertStrTo<double>(value);
+		const double v = mpt::parse<double>(value);
 
 		const bool isTime = SFZStartsWith(key, "time"), isLevel = SFZStartsWith(key, "level");
 		std::string_view pointStr;
@@ -82,7 +91,7 @@ struct SFZFlexEG
 
 		if(!pointStr.empty() && SFZIsNumeric(pointStr))
 		{
-			PointIndex point = ConvertStrTo<PointIndex>(std::string(pointStr));
+			PointIndex point = mpt::parse<PointIndex>(std::string(pointStr));
 			if(point >= points.size() && point < MAX_ENVPOINTS)
 				points.resize(point + 1);
 
@@ -159,7 +168,7 @@ struct SFZFlexEG
 				break;
 		}
 
-		if(sustain < env.size())
+		if(sustain < env.size() && !(envType == ENV_VOLUME && sustain == env.size() - 1u && env.back().value == 0))
 		{
 			env.nSustainStart = env.nSustainEnd = sustain;
 			env.dwFlags.set(ENV_SUSTAIN);
@@ -181,16 +190,16 @@ protected:
 
 	static EnvelopeNode::value_t ToValue(double value, double scale, double minVal, double maxVal, const std::function<double(double)> &conversionFunc)
 	{
-		value = conversionFunc((value * scale - minVal) / (maxVal - minVal)) * ENVELOPE_MAX + ENVELOPE_MIN;
-		Limit<double, double>(value, ENVELOPE_MIN, ENVELOPE_MAX);
+		value = conversionFunc((value * scale - minVal) / (maxVal - minVal)) * double(ENVELOPE_MAX) + double(ENVELOPE_MIN);
+		Limit(value, double(ENVELOPE_MIN), double(ENVELOPE_MAX));
 		return mpt::saturate_round<EnvelopeNode::value_t>(value);
 	}
 
 	static double Identity(double v) noexcept { return v; }
 
-	static double CentsToFilterCutoff(double v, const CSoundFile &sndFile, int envBaseCutoff, uint32 envBaseFreq)
+	static double CentsToFilterCutoff(double v, const CSoundFile &sndFile, int envBaseCutoff, float envBaseFreq)
 	{
-		const auto freq = envBaseFreq * std::pow(2.0, v / 1200.0);
+		const auto freq = static_cast<double>(envBaseFreq) * std::pow(2.0, v / 1200.0);
 		return Util::muldivr(sndFile.FrequencyToCutOff(freq), 127, envBaseCutoff) / 127.0;
 	}
 
@@ -210,7 +219,7 @@ struct SFZEnvelope
 	void Parse(std::string_view key, const std::string &value)
 	{
 		key = key.substr(key.find('_') + 1);
-		double v = ConvertStrTo<double>(value);
+		double v = mpt::parse<double>(value);
 		if(key == "depth")
 			Limit(v, -12000.0, 12000.0);
 		else if(key == "start" || key == "sustain")
@@ -298,7 +307,7 @@ struct SFZRegion
 	};
 
 	size_t filenameOffset = 0;
-	std::string filename, name;
+	std::string filename, globalName, regionName;
 	SFZEnvelope ampEnv, pitchEnv, filterEnv;
 	std::vector<SFZFlexEG> flexEGs;
 	SmpLength loopStart = 0, loopEnd = 0;
@@ -327,7 +336,7 @@ struct SFZRegion
 	template<typename T, typename Tc>
 	static void Read(const std::string &valueStr, T &value, Tc valueMin = std::numeric_limits<T>::min(), Tc valueMax = std::numeric_limits<T>::max())
 	{
-		double valueF = ConvertStrTo<double>(valueStr);
+		double valueF = mpt::parse<double>(valueStr);
 		if constexpr(std::numeric_limits<T>::is_integer)
 		{
 			valueF = mpt::round(valueF);
@@ -345,7 +354,7 @@ struct SFZRegion
 		if(value[0] >= '0' && value[0] <= '9')
 		{
 			// MIDI key
-			key = ConvertStrTo<uint8>(value);
+			key = mpt::parse<uint8>(value);
 		} else if(value.length() < 2)
 		{
 			return 0;
@@ -375,7 +384,7 @@ struct SFZRegion
 			if(octaveOffset >= value.length())
 				return 0;
 
-			int8 octave = ConvertStrTo<int8>(value.c_str() + octaveOffset);
+			int8 octave = mpt::parse<int8>(value.c_str() + octaveOffset);
 			key += (octave + 1) * 12;
 		}
 		key += control.octaveOffset * 12 + control.noteOffset;
@@ -389,8 +398,10 @@ struct SFZRegion
 			filename = control.defaultPath + value;
 			filenameOffset = control.defaultPath.size();
 		}
+		else if(key == "global_label")
+			globalName = value;
 		else if(key == "region_label")
-			name = value;
+			regionName = value;
 		else if(key == "lokey")
 			keyLo = ReadKey(value, control);
 		else if(key == "hikey")
@@ -479,7 +490,7 @@ struct SFZRegion
 			pitchEnv.Parse(key, value);
 		else if(SFZStartsWith(key, "eg") && SFZIsNumeric(key.substr(2, 2)) && key.substr(4, 1) == "_")
 		{
-			uint8 eg = ConvertStrTo<uint8>(std::string(key.substr(2, 2)));
+			uint8 eg = mpt::parse<uint8>(std::string(key.substr(2, 2)));
 			if(eg >= flexEGs.size())
 				flexEGs.resize(eg + 1);
 			flexEGs[eg].Parse(key, value);
@@ -490,10 +501,10 @@ struct SFZRegion
 struct SFZInputFile
 {
 	FileReader file;
-	std::unique_ptr<InputFile> inputFile;  // FileReader has pointers into this so its address must not change
+	std::unique_ptr<mpt::IO::InputFile> inputFile;  // FileReader has pointers into this so its address must not change
 	std::string remain;
 
-	SFZInputFile(FileReader f = {}, std::unique_ptr<InputFile> i = {}, std::string r = {})
+	SFZInputFile(FileReader f = {}, std::unique_ptr<mpt::IO::InputFile> i = {}, std::string r = {})
 		: file{std::move(f)}, inputFile{std::move(i)}, remain{std::move(r)} {}
 	SFZInputFile(SFZInputFile &&) = default;
 };
@@ -646,12 +657,12 @@ bool CSoundFile::ReadSFZInstrument(INSTRUMENTINDEX nInstr, FileReader &file)
 				if(!filename.empty())
 				{
 					if(filenameU8.find(':') == std::string::npos)
-						filename = file.GetFileName().GetPath() + filename;
+						filename = file.GetOptionalFileName().value_or(P_("")).GetDirectoryWithDrive() + filename;
 					filename = filename.Simplify();
 					// Avoid recursive #include
-					if(std::find_if(files.begin(), files.end(), [&filename](const SFZInputFile &f) { return f.file.GetFileName() == filename; }) == files.end())
+					if(std::find_if(files.begin(), files.end(), [&filename](const SFZInputFile &f) { return f.file.GetOptionalFileName().value_or(P_("")) == filename; }) == files.end())
 					{
-						auto f = std::make_unique<InputFile>(filename);
+						auto f = std::make_unique<mpt::IO::InputFile>(filename);
 						if(f->IsValid())
 						{
 							s.erase(0, charsRead);
@@ -752,6 +763,7 @@ bool CSoundFile::ReadSFZInstrument(INSTRUMENTINDEX nInstr, FileReader &file)
 	DestroyInstrument(nInstr, deleteAssociatedSamples);
 	if(nInstr > m_nInstruments) m_nInstruments = nInstr;
 	Instruments[nInstr] = pIns;
+	pIns->name = mpt::ToCharset(GetCharsetInternal(), mpt::Charset::UTF8, globals.globalName);
 
 	SAMPLEINDEX prevSmp = 0;
 	for(auto &region : regions)
@@ -775,7 +787,7 @@ bool CSoundFile::ReadSFZInstrument(INSTRUMENTINDEX nInstr, FileReader &file)
 			sample.uFlags.set(CHN_16BIT);
 			std::function<uint16(int32)> generator;
 			if(synthSample == "*sine")
-				generator = [](int32 i) { return mpt::saturate_round<int16>(std::sin(i * 2.0 * M_PI / 256.0) * int16_max); };
+				generator = [](int32 i) { return mpt::saturate_round<int16>(std::sin(i * ((2.0 * mpt::numbers::pi) / 256.0)) * int16_max); };
 			else if(synthSample == "*square")
 				generator = [](int32 i) { return i < 128 ? int16_max : int16_min; };
 			else if(synthSample == "*triangle" || synthSample == "*tri")
@@ -813,11 +825,11 @@ bool CSoundFile::ReadSFZInstrument(INSTRUMENTINDEX nInstr, FileReader &file)
 		{
 			if(region.filename.find(':') == std::string::npos)
 			{
-				filename = file.GetFileName().GetPath() + filename;
+				filename = file.GetOptionalFileName().value_or(P_("")).GetDirectoryWithDrive() + filename;
 			}
 			filename = filename.Simplify();
 			SetSamplePath(smp, filename);
-			InputFile f(filename, SettingCacheCompleteFileBeforeLoading());
+			mpt::IO::InputFile f(filename, SettingCacheCompleteFileBeforeLoading());
 			FileReader smpFile = GetFileReader(f);
 			if(!ReadSampleFromFile(smp, smpFile, false))
 			{
@@ -832,10 +844,10 @@ bool CSoundFile::ReadSFZInstrument(INSTRUMENTINDEX nInstr, FileReader &file)
 			sample.uFlags.set(SMP_KEEPONDISK, sample.HasSampleData());
 		}
 
-		if(!region.name.empty())
-			m_szNames[smp] = mpt::ToCharset(GetCharsetInternal(), mpt::Charset::UTF8, region.name);
+		if(!region.regionName.empty())
+			m_szNames[smp] = mpt::ToCharset(GetCharsetInternal(), mpt::Charset::UTF8, region.regionName);
 		if(!m_szNames[smp][0])
-			m_szNames[smp] = mpt::ToCharset(GetCharsetInternal(), mpt::PathString::FromUTF8(region.filename).GetFileName().ToUnicode());
+			m_szNames[smp] = mpt::ToCharset(GetCharsetInternal(), mpt::PathString::FromUTF8(region.filename).GetFilenameBase().ToUnicode());
 
 		if(region.useSampleKeyRoot)
 		{
@@ -864,11 +876,11 @@ bool CSoundFile::ReadSFZInstrument(INSTRUMENTINDEX nInstr, FileReader &file)
 		pIns->nCutSwing = mpt::saturate_round<uint8>(region.filterRandom * (m_SongFlags[SONG_EXFILTERRANGE] ? 20 : 24) / 1200.0);
 		pIns->midiPWD = mpt::saturate_round<int8>(region.pitchBend / 100.0);
 
-		pIns->nNNA = NNA_NOTEOFF;
+		pIns->nNNA = NewNoteAction::NoteOff;
 		if(region.polyphony == 1)
 		{
-			pIns->nDNA = DNA_NOTECUT;
-			pIns->nDCT = DCT_SAMPLE;
+			pIns->nDNA = DuplicateNoteAction::NoteCut;
+			pIns->nDCT = DuplicateCheckType::Sample;
 		}
 		region.ampEnv.ConvertToMPT(pIns, *this, ENV_VOLUME);
 		if(region.pitchEnv.depth)
@@ -912,13 +924,13 @@ bool CSoundFile::ReadSFZInstrument(INSTRUMENTINDEX nInstr, FileReader &file)
 			switch(region.loopMode)
 			{
 			case SFZRegion::LoopMode::kContinuous:
-			case SFZRegion::LoopMode::kOneShot:
 				sample.uFlags.set(CHN_LOOP);
 				break;
 			case SFZRegion::LoopMode::kSustain:
 				sample.uFlags.set(CHN_SUSTAINLOOP);
 				break;
 			case SFZRegion::LoopMode::kNoLoop:
+			case SFZRegion::LoopMode::kOneShot:
 				sample.uFlags.reset(CHN_LOOP | CHN_SUSTAINLOOP);
 			}
 		}
@@ -1003,7 +1015,7 @@ bool CSoundFile::ReadSFZInstrument(INSTRUMENTINDEX nInstr, FileReader &file)
 			// Avoid ruining the original samples
 			if(auto filename = GetSamplePath(smp); !filename.empty())
 			{
-				filename = filename.GetPath() + filename.GetFileName() + filenameModifier + filename.GetFileExt();
+				filename = filename.GetDirectoryWithDrive() + filename.GetFilenameBase() + filenameModifier + filename.GetFilenameExtension();
 				SetSamplePath(smp, filename);
 			}
 		}
@@ -1031,7 +1043,7 @@ static void WriteSFZEnvelope(std::ostream &f, double tickDuration, int index, co
 
 	const bool sustainAtEnd = (!env.dwFlags[ENV_SUSTAIN] || env.nSustainStart == (env.size() - 1)) && convFunc(env.back().value) != 0.0;
 
-	const auto prefix = mpt::format("\neg%1_")(mpt::fmt::dec0<2>(index));
+	const auto prefix = MPT_AFORMAT("\neg{}_")(mpt::afmt::dec0<2>(index));
 	f << "\n" << prefix << type << "=" << scale;
 	f << prefix << "points=" << (env.size() + (sustainAtEnd ? 1 : 0));
 	EnvelopeNode::tick_t lastTick = 0;
@@ -1061,12 +1073,25 @@ static void WriteSFZEnvelope(std::ostream &f, double tickDuration, int index, co
 		f << "\n// Release Node: " << static_cast<uint32>(env.nReleaseNode);
 }
 
+static std::string SanitizeSFZString(std::string s, mpt::Charset sourceCharset)
+{
+	using namespace std::literals;
+	// Remove characters could trip up the parser
+	std::string::size_type pos = 0;
+	while((pos = s.find_first_of("<=\r\n\t\0"sv, pos)) != std::string::npos)
+	{
+		s[pos++] = ' ';
+	}
+	return mpt::ToCharset(mpt::Charset::UTF8, sourceCharset, s);
+}
+
+
 bool CSoundFile::SaveSFZInstrument(INSTRUMENTINDEX nInstr, std::ostream &f, const mpt::PathString &filename, bool useFLACsamples) const
 {
 #ifdef MODPLUG_TRACKER
-	const mpt::FlushMode flushMode = mpt::FlushModeFromBool(TrackerSettings::Instance().MiscFlushFileBuffersOnSave);
+	const mpt::IO::FlushMode flushMode = mpt::IO::FlushModeFromBool(TrackerSettings::Instance().MiscFlushFileBuffersOnSave);
 #else
-	const mpt::FlushMode flushMode = mpt::FlushMode::Full;
+	const mpt::IO::FlushMode flushMode = mpt::IO::FlushMode::Full;
 #endif
 	const ModInstrument *ins = Instruments[nInstr];
 	if(ins == nullptr)
@@ -1074,30 +1099,26 @@ bool CSoundFile::SaveSFZInstrument(INSTRUMENTINDEX nInstr, std::ostream &f, cons
 
 	// Creating directory names with trailing spaces or dots is a bad idea, as they are difficult to remove in Windows.
 	const mpt::RawPathString whitespaceDirName = PL_(" \n\r\t.");
-	const mpt::PathString sampleBaseName = mpt::PathString::FromNative(mpt::String::Trim(filename.GetFileName().AsNative(), whitespaceDirName));
+	const mpt::PathString sampleBaseName = mpt::PathString::FromNative(mpt::trim(filename.GetFilenameBase().AsNative(), whitespaceDirName));
 	const mpt::PathString sampleDirName = (sampleBaseName.empty() ? P_("Samples") : sampleBaseName)  + P_("/");
-	const mpt::PathString sampleBasePath = filename.GetPath() + sampleDirName;
-	if(!sampleBasePath.IsDirectory() && !::CreateDirectory(sampleBasePath.AsNative().c_str(), nullptr))
+	const mpt::PathString sampleBasePath = filename.GetDirectoryWithDrive() + sampleDirName;
+	if(!mpt::native_fs{}.is_directory(sampleBasePath) && !::CreateDirectory(sampleBasePath.AsNative().c_str(), nullptr))
 		return false;
 
 	const double tickDuration = m_PlayState.m_nSamplesPerTick / static_cast<double>(m_MixerSettings.gdwMixingFreq);
 
 	f << std::setprecision(10);
-	if(!ins->name.empty())
-	{
-		f << "// Name: " << mpt::ToCharset(mpt::Charset::UTF8, GetCharsetInternal(), ins->name) << "\n";
-	}
 	f << "// Created with " << mpt::ToCharset(mpt::Charset::UTF8, Version::Current().GetOpenMPTVersionString()) << "\n";
 	f << "// Envelope tempo base: tempo " << m_PlayState.m_nMusicTempo.ToDouble();
 	switch(m_nTempoMode)
 	{
-	case tempoModeClassic:
+	case TempoMode::Classic:
 		f << " (classic tempo mode)";
 		break;
-	case tempoModeAlternative:
+	case TempoMode::Alternative:
 		f << " (alternative tempo mode)";
 		break;
-	case tempoModeModern:
+	case TempoMode::Modern:
 		f << ", " << m_PlayState.m_nMusicSpeed << " ticks per row, " << m_PlayState.m_nCurrentRowsPerBeat << " rows per beat (modern tempo mode)";
 		break;
 	default:
@@ -1105,8 +1126,12 @@ bool CSoundFile::SaveSFZInstrument(INSTRUMENTINDEX nInstr, std::ostream &f, cons
 		break;
 	}
 
-	f << "\n\n<control>\ndefault_path=" << sampleDirName.ToUTF8() << "\n\n";
-	f << "<group>";
+	f << "\n\n<control>\ndefault_path=" << sampleDirName.ToUTF8();
+	if(const auto globalName = SanitizeSFZString(ins->name, GetCharsetInternal()); !globalName.empty())
+	{
+		f << "\n\n<global>\nglobal_label=" << globalName;
+	}
+	f << "\n\n<group>";
 	f << "\nbend_up=" << ins->midiPWD * 100;
 	f << "\nbend_down=" << -ins->midiPWD * 100;
 	const uint32 cutoff = ins->IsCutoffEnabled() ? ins->GetCutoff() : 127;
@@ -1127,19 +1152,19 @@ bool CSoundFile::SaveSFZInstrument(INSTRUMENTINDEX nInstr, std::ostream &f, cons
 		f << "\nampeg_release_shape=0";
 	}
 
-	if(ins->nDNA == DNA_NOTECUT && ins->nDCT != DCT_NONE)
+	if(ins->nDNA == DuplicateNoteAction::NoteCut && ins->nDCT != DuplicateCheckType::None)
 		f << "\npolyphony=1";
 
 	WriteSFZEnvelope(f, tickDuration, 1, ins->VolEnv, "amplitude", 100.0, [](int32 val) { return val / static_cast<double>(ENVELOPE_MAX); });
 	WriteSFZEnvelope(f, tickDuration, 2, ins->PanEnv, "pan", 100.0, [](int32 val) { return 2.0 * (val - ENVELOPE_MID) / (ENVELOPE_MAX - ENVELOPE_MIN); });
 	if(ins->PitchEnv.dwFlags[ENV_FILTER])
 	{
-		const auto envScale = 1200.0 * std::log(CutOffToFrequency(127, 256) / static_cast<double>(CutOffToFrequency(0, -256))) / M_LN2;
+		const auto envScale = 1200.0 * std::log(CutOffToFrequency(127, 256) / static_cast<double>(CutOffToFrequency(0, -256))) / mpt::numbers::ln2;
 		const auto cutoffNormal = CutOffToFrequency(cutoff);
 		WriteSFZEnvelope(f, tickDuration, 3, ins->PitchEnv, "cutoff", envScale, [this, cutoff, cutoffNormal, envScale](int32 val) {
 			// Convert interval between center frequency and envelope into cents
 			const auto freq = CutOffToFrequency(cutoff, (val - ENVELOPE_MID) * 256 / (ENVELOPE_MAX - ENVELOPE_MID));
-			return 1200.0 * std::log(freq / static_cast<double>(cutoffNormal)) / M_LN2 / envScale;
+			return 1200.0 * std::log(freq / static_cast<double>(cutoffNormal)) / mpt::numbers::ln2 / envScale;
 		});
 	} else
 	{
@@ -1179,31 +1204,38 @@ bool CSoundFile::SaveSFZInstrument(INSTRUMENTINDEX nInstr, std::ostream &f, cons
 		else
 			sampleName += P_(".wav");
 
+		bool success = false;
 		try
 		{
-			mpt::SafeOutputFile fSmp(sampleName, std::ios::binary, flushMode);
-			if(fSmp)
+			mpt::IO::SafeOutputFile sfSmp(sampleName, std::ios::binary, flushMode);
+			if(sfSmp)
 			{
-				//fSmp.exceptions(fSmp.exceptions() | std::ios::badbit | std::ios::failbit);
+				mpt::IO::ofstream &fSmp = sfSmp;
+				fSmp.exceptions(fSmp.exceptions() | std::ios::badbit | std::ios::failbit);
 
 				if(isAdlib)
-					SaveS3ISample(ins->Keyboard[i], fSmp);
+					success = SaveS3ISample(ins->Keyboard[i], fSmp);
 				else if(useFLACsamples)
-					SaveFLACSample(ins->Keyboard[i], fSmp);
+					success = SaveFLACSample(ins->Keyboard[i], fSmp);
 				else
-					SaveWAVSample(ins->Keyboard[i], fSmp);
+					success = SaveWAVSample(ins->Keyboard[i], fSmp);
 			}
 		} catch(const std::exception &)
+		{
+			success = false;
+		}
+		if(!success)
 		{
 			AddToLog(LogError, MPT_USTRING("Unable to save sample: ") + sampleName.ToUnicode());
 		}
 
+
 		f << "\n\n<region>";
-		if(!m_szNames[ins->Keyboard[i]].empty())
+		if(const auto regionName = SanitizeSFZString(m_szNames[ins->Keyboard[i]], GetCharsetInternal()); !regionName.empty())
 		{
-			f << "\nregion_label=" << mpt::ToCharset(mpt::Charset::UTF8, GetCharsetInternal(), m_szNames[ins->Keyboard[i]]);
+			f << "\nregion_label=" << regionName;
 		}
-		f << "\nsample=" << sampleName.GetFullFileName().ToUTF8();
+		f << "\nsample=" << sampleName.GetFilename().ToUTF8();
 		f << "\nlokey=" << i;
 		f << "\nhikey=" << endOfRegion;
 		if(sample.rootNote != NOTE_NONE)
